import re
from string import ascii_lowercase, ascii_uppercase
from unittest.mock import MagicMock, PropertyMock

import pytest

from shrinkray.passes.definitions import ParseError
from shrinkray.passes.genericlanguages import (
    IntegerFormat,
    RegionReplacingPatches,
    Substring,
    combine_expressions,
    cut_comment_like_things,
    cut_comments,
    iter_indices,
    merge_adjacent_strings,
    normalize_identifiers,
    reduce_integer_literals,
    regex_pass,
    replace_falsey_with_zero,
    shortlex,
    simplify_brackets,
)
from shrinkray.work import WorkContext
from tests.helpers import reduce_with


def test_can_reduce_an_integer_in_the_middle_of_a_string() -> None:
    assert (
        reduce_with([reduce_integer_literals], b"bobcats99999hello", lambda x: True)
        == b"bobcats0hello"
    )


def test_can_reduce_integers_to_boundaries() -> None:
    assert (
        reduce_with([reduce_integer_literals], b"100", lambda x: eval(x) >= 73) == b"73"
    )


def test_can_combine_expressions() -> None:
    assert reduce_with([combine_expressions], b"10 + 10", lambda x: True) == b"20"


def test_does_not_error_on_bad_expression() -> None:
    assert reduce_with([combine_expressions], b"1 / 0", lambda x: True) == b"1 / 0"


def test_can_combine_expressions_with_no_expressions() -> None:
    assert (
        reduce_with([combine_expressions], b"hello world", lambda x: True)
        == b"hello world"
    )


LOTS_OF_COMMENTS = b"""
hello # this is a single line comment
/* this
is
a
multiline
comment */ world // some extraneous garbage
"""


def test_comment_removal():
    x = reduce_with([cut_comment_like_things], LOTS_OF_COMMENTS, lambda x: True)
    lines = [line.strip() for line in x.splitlines() if line.strip()]
    assert lines == [b"hello", b"world"]


# === Substring tests ===


def test_substring_parse_valid():
    sub = Substring(prefix=b"<", suffix=b">")
    assert sub.parse(b"<hello>") == b"hello"


def test_substring_parse_invalid_prefix():
    sub = Substring(prefix=b"<", suffix=b">")
    with pytest.raises(ParseError):
        sub.parse(b"hello>")


def test_substring_parse_invalid_suffix():
    sub = Substring(prefix=b"<", suffix=b">")
    with pytest.raises(ParseError):
        sub.parse(b"<hello")


def test_substring_dumps():
    sub = Substring(prefix=b"<", suffix=b">")
    assert sub.dumps(b"hello") == b"<hello>"


def test_substring_name():
    sub = Substring(prefix=b"<", suffix=b">")
    assert sub.name == "Substring(1, 1)"


# === RegionReplacingPatches tests ===


def test_region_replacing_patches_size():
    patches = RegionReplacingPatches([(0, 5)])
    # Replacing 5 chars with 3 chars = size savings of 2
    assert patches.size({0: b"abc"}) == 2


def test_region_replacing_patches_size_empty_patch_raises():
    patches = RegionReplacingPatches([(0, 5)])
    with pytest.raises(AssertionError):
        patches.size({})


# === IntegerFormat tests ===


def test_integer_format_parse_valid():
    fmt = IntegerFormat()
    assert fmt.parse(b"123") == 123


def test_integer_format_parse_invalid():
    fmt = IntegerFormat()
    with pytest.raises(ParseError):
        fmt.parse(b"abc")


def test_integer_format_parse_invalid_unicode():
    fmt = IntegerFormat()
    with pytest.raises(ParseError):
        fmt.parse(b"\xff\xfe")


def test_integer_format_dumps():
    fmt = IntegerFormat()
    assert fmt.dumps(42) == b"42"


# === shortlex tests ===


def test_shortlex_shorter_wins():
    assert shortlex(b"ab") < shortlex(b"abc")


def test_shortlex_same_length_lex_order():
    assert shortlex(b"ab") < shortlex(b"ba")


# === iter_indices tests ===


def test_iter_indices_finds_all():
    result = list(iter_indices(b"abcabc", b"abc"))
    assert result == [0, 3]


def test_iter_indices_no_match():
    result = list(iter_indices(b"hello", b"xyz"))
    assert result == []


def test_iter_indices_single_match():
    result = list(iter_indices(b"hello", b"ell"))
    assert result == [1]


# === merge_adjacent_strings tests ===


def test_merge_adjacent_strings():
    result = reduce_with([merge_adjacent_strings], b"'' ''", lambda x: True)
    assert result == b" "


def test_merge_adjacent_double_quotes():
    result = reduce_with([merge_adjacent_strings], b'"" ""', lambda x: True)
    assert result == b" "


# === replace_falsey_with_zero tests ===


def test_replace_falsey_empty_string():
    result = reduce_with([replace_falsey_with_zero], b"''", lambda x: True)
    assert result == b"0"


def test_replace_falsey_false():
    result = reduce_with([replace_falsey_with_zero], b"false", lambda x: True)
    assert result == b"0"


def test_replace_falsey_empty_list():
    result = reduce_with([replace_falsey_with_zero], b"[]", lambda x: True)
    assert result == b"0"


def test_replace_falsey_empty_parens():
    result = reduce_with([replace_falsey_with_zero], b"()", lambda x: True)
    assert result == b"0"


# === simplify_brackets tests ===


def test_simplify_brackets_curly_to_square():
    result = reduce_with([simplify_brackets], b"{}", lambda x: True)
    assert result == b"()"


def test_simplify_brackets_square_to_parens():
    result = reduce_with([simplify_brackets], b"[]", lambda x: True)
    assert result == b"()"


# === normalize_identifiers tests ===


def test_normalize_identifiers_basic():
    result = reduce_with([normalize_identifiers], b"hello hello", lambda x: True)
    # Tries uppercase first, then lowercase
    assert result == b"A A"


def test_normalize_identifiers_multiple():
    result = reduce_with([normalize_identifiers], b"foo bar foo bar", lambda x: True)
    # Should replace with shorter names
    assert len(result) < len(b"foo bar foo bar")


def test_normalize_identifiers_target_not_found():
    """Test the continue branch when target is not in current source.

    When a target identifier is no longer in the source after a previous replacement.
    """
    result = reduce_with(
        [normalize_identifiers],
        b"abc ab abc",  # Both abc and ab are identifiers
        lambda x: b"ab" in x,  # Keep any version with "ab" in it
    )
    # "abc" should be replaceable since it's larger and "ab" is preserved
    assert b"ab" in result


def test_normalize_identifiers_identifier_disappears():
    """Test when an identifier disappears after a replacement.

    When 'hello' is replaced with 'A', the identifier 'hello' no longer exists,
    so when we try to find 'hello' again in a later iteration it won't be found.
    """
    # Create a situation where one identifier gets replaced and then
    # we look for another identifier that doesn't exist in the new source
    result = reduce_with(
        [normalize_identifiers],
        b"longname x longname",  # longname and x are identifiers
        lambda x: x.count(b" ") == 2,  # Keep 2 spaces
    )
    # longname should be replaced with something shorter
    assert b"longname" not in result


def test_regex_pass_with_compiled_pattern():
    """Test regex_pass with an already compiled pattern."""

    # Create a pass with a pre-compiled pattern
    pattern = re.compile(rb"[0-9]+")

    @regex_pass(pattern)
    async def reduce_numbers(problem):
        await problem.is_interesting(b"0")

    result = reduce_with([reduce_numbers], b"abc123def", lambda x: True)
    assert result == b"abc0def"


def test_reduce_integer_early_termination():
    """Test reduce_integer when lo + 1 is interesting.

    The reduce_integer function does binary search with special handling
    for early termination when lo + 1 is interesting. This happens when
    the search narrows down to where lo + 1 equals the minimum interesting value.
    """
    result = reduce_with(
        [reduce_integer_literals],
        b"100",
        lambda x: int(x) >= 5,  # Only values >= 5 are interesting
    )
    assert result == b"5"


def test_cut_comments_no_end_found():
    """Test cut_comments when include_end=True but end marker not found."""

    # Test with include_end=True and no end marker
    result = reduce_with(
        [lambda p: cut_comments(p, b"/*", b"*/", include_end=True)],
        b"hello /* comment without end",
        lambda x: True,
    )
    # Comment should not be removed since end marker wasn't found
    assert b"/*" in result


def test_cut_comments_no_end_include_false():
    """Test cut_comments when include_end=False and end not found."""

    # Test with include_end=False and no end marker - should cut to end
    result = reduce_with(
        [lambda p: cut_comments(p, b"#", b"\n", include_end=False)],
        b"hello # comment to end",  # No newline at end
        lambda x: True,
    )
    # Comment should be removed to end of string
    assert result == b"hello "


def test_normalize_identifiers_all_letters_used():
    """Test when all single-letter identifiers are already used.

    When all lowercase and uppercase letters exist as identifiers,
    the inner loop completes without breaking, triggering the 221->220 branch.
    """
    # Create a source with all 26 lowercase and 26 uppercase letters as identifiers

    all_letters = " ".join(ascii_lowercase + ascii_uppercase).encode("ascii")
    result = reduce_with(
        [normalize_identifiers],
        all_letters,
        lambda x: True,
    )
    # Result should still be valid - letters might be replaced with numbers etc
    assert len(result) <= len(all_letters)


async def test_normalize_identifiers_target_disappears():
    """Test when an identifier disappears before its turn in the loop.

    Exercises the defensive continue statement when an identifier that was
    in the original source is no longer present when we try to process it.
    This can happen in concurrent scenarios or with modified problem classes.
    """

    # Create a mock problem that changes its current_test_case
    mock_problem = MagicMock()
    mock_problem.work = WorkContext(parallelism=1)

    # Track call count to simulate changing source
    # Access pattern:
    # 1. Line 217: get identifiers from initial source
    # 2. Line 233 (first iter): get source for first target (longer one)
    # 3. Line 233 (second iter): get source for second target - should NOT contain it
    access_count = [0]
    original_source = b"aaa bb"  # 'aaa' is longer, processed first

    def get_current_test_case():
        access_count[0] += 1
        # After first two accesses, return source without 'bb'
        # Access 1: identifiers (contains both aaa, bb)
        # Access 2: first iteration source check (for 'aaa')
        # Access 3+: return source without 'bb' to trigger continue
        if access_count[0] > 2:
            return b"aaa only"  # 'bb' is gone but 'aaa' remains
        return original_source

    type(mock_problem).current_test_case = PropertyMock(
        side_effect=get_current_test_case
    )

    # Make is_interesting always return False so no replacements happen
    async def mock_is_interesting(x):
        return False

    mock_problem.is_interesting = mock_is_interesting

    # This should hit the continue branch when 'bb' can't be found
    await normalize_identifiers(mock_problem)



import os
from glob import glob

import libcst
import libcst.matchers as m
import pytest

from shrinkray.passes.python import (
    PYTHON_PASSES,
    is_python,
    libcst_transform,
    lift_indented_constructs,
    replace_bodies_with_ellipsis,
    replace_statements_with_pass,
    strip_annotations,
)
from shrinkray.problem import BasicReductionProblem
from shrinkray.work import WorkContext
from tests.helpers import reduce, reduce_with


def test_can_replace_blocks_with_body() -> None:
    body = reduce_with(
        [lift_indented_constructs], b"if True:\n    x = 1", lambda t: b"x" in t
    )
    assert body == b"x = 1"


ELIF_BLOCK = b"""
if True:
    x = 1
elif True:
    x = 2
"""


def test_lifts_bodies_of_elif():
    assert (
        reduce_with([lift_indented_constructs], ELIF_BLOCK, lambda x: True).strip()
        == b"x = 1"
    )


def test_does_not_error_on_elif():
    assert (
        reduce_with([lift_indented_constructs], ELIF_BLOCK, lambda x: b"elif" in x)
        == ELIF_BLOCK
    )


def test_lifts_bodies_of_with():
    assert (
        reduce_with(
            [lift_indented_constructs], b"with ...:\n    x = 1", lambda x: True
        ).strip()
        == b"x = 1"
    )


def test_can_replace_statements_with_pass() -> None:
    body = reduce_with(
        [replace_statements_with_pass], b"from x import *", lambda t: len(t) > 1
    )
    assert body == b"pass"


def test_can_reduce_an_example_that_crashes_lib_cst():
    reduce(b"() if 0 else(lambda:())", lambda x: len(x) >= 5)


ANNOTATED = b"""
def has_an_annotation(x: list[int]) -> list[int]:
    y: list[int] = list(reversed(x))
    return x + y
"""

DEANNOTATED = b"""
def has_an_annotation(x):
    y = list(reversed(x))
    return x + y
"""


def test_strip_annotations():
    assert reduce_with([strip_annotations], ANNOTATED, lambda x: True) == DEANNOTATED


def test_single_annotation():
    x = b"x:A\n"
    assert reduce_with(PYTHON_PASSES, x, lambda y: True).strip() == b""


IF_BLOCK = b"""
if True:
    x = 1
    y = 2
    assert x + y
"""


def test_body_replacement_of_if():
    assert (
        reduce_with([replace_bodies_with_ellipsis], IF_BLOCK, lambda x: True).strip()
        == b"if True:\n    ..."
    )


ROOT = os.path.dirname(os.path.dirname(__file__))


PYTHON_FILES = glob(os.path.join(ROOT, "src", "**", "*.py"), recursive=True) + glob(
    os.path.join(ROOT, "tests", "**", "*.py"), recursive=True
)


ISSUE_12_INPUT = b"""
import asyncio
import _lsprof

if True:
    a = 1
    b = 2
    c = 3

if True:
    obj = _lsprof.Profiler()
    obj.enable()
    obj._pystart_callback(lambda: 0, 0)
    obj = None
    loop = asyncio.get_event_loop()
"""

ISSUE_12_OUTPUT = b"""
import asyncio
import _lsprof

if True:
    ...

if True:
    obj = _lsprof.Profiler()
    obj.enable()
    obj._pystart_callback(lambda: 0, 0)
    obj = None
    loop = asyncio.get_event_loop()
"""


def test_reduce_with_ellipsis_can_reduce_single_block():
    reduced = reduce_with(
        [replace_bodies_with_ellipsis],
        ISSUE_12_INPUT,
        lambda x: b"Profiler" in x,
        parallelism=1,
    )
    assert reduced == ISSUE_12_OUTPUT


def test_python_passes_handle_invalid_syntax():
    """Test that PYTHON_PASSES handle invalid Python syntax gracefully."""
    invalid_python = b"def foo(:\n    pass"  # Missing parameter

    # Should not raise, just return the original
    result = reduce_with(PYTHON_PASSES, invalid_python, lambda x: True)
    assert result == invalid_python


def test_python_passes_handle_non_python():
    """Test that PYTHON_PASSES handle non-Python content gracefully."""
    not_python = b"This is not Python code at all {{{}}}"

    # Should not raise, just return the original
    result = reduce_with(PYTHON_PASSES, not_python, lambda x: True)
    assert result == not_python


def test_libcst_validation_error_handling():
    """Test that CSTValidationError is handled gracefully.

    The Python passes can encounter situations where a transformation
    would create an invalid CST structure. These should be caught
    and handled gracefully.
    """
    # Code that when reduced might try to create invalid structures
    # This exercises the CSTValidationError exception handling path
    code = b"""
class A:
    def method(self):
        pass
"""
    # Force various reductions to exercise validation paths
    result = reduce_with(PYTHON_PASSES, code, lambda x: len(x) > 10)
    assert len(result) <= len(code)


def test_python_reduction_with_complex_annotations():
    """Test reduction of code with complex type annotations."""
    code = b"""
def complex_function(
    x: list[tuple[int, str]],
    y: dict[str, list[int]] | None = None
) -> tuple[int, ...]:
    z: set[int] = set()
    return (1, 2, 3)
"""
    result = reduce_with([strip_annotations], code, lambda x: b"def" in x)
    # Annotations should be stripped
    assert b": list[tuple" not in result
    assert b"-> tuple" not in result


def test_is_python_returns_true_for_valid_python():
    """Test that is_python returns True for valid Python code."""
    valid_code = b"x = 1"
    assert is_python(valid_code) is True


def test_is_python_returns_false_for_invalid_python():
    """Test that is_python returns False for invalid Python code."""
    invalid_code = b"def foo(:"
    assert is_python(invalid_code) is False


async def test_libcst_transform_handles_test_case_becoming_invalid():
    """Test that libcst_transform handles the case where the test case becomes invalid Python mid-reduction.

    The test case is valid initially but becomes invalid between calls, which
    should trigger ParserSyntaxError handling in can_apply.
    """
    # Track how many times current_test_case is accessed
    access_count = [0]
    valid_code = b"x = 1\ny = 2\nz = 3\n"
    invalid_code = b"def foo(:\n"

    class MockProblem:
        def __init__(self):
            self.work = WorkContext(parallelism=1)
            self._test_case = valid_code

        @property
        def current_test_case(self):
            access_count[0] += 1
            # First access returns valid code (for initial parse)
            # Later accesses return invalid code (to trigger ParserSyntaxError in can_apply)
            if access_count[0] <= 2:
                return valid_code
            return invalid_code

        def sort_key(self, x):
            return (len(x), x)

        async def is_interesting(self, x):
            return True

    problem = MockProblem()

    await libcst_transform(
        problem,  # type: ignore
        m.SimpleStatementLine(),
        lambda x: libcst.RemoveFromParent(),
    )
    # The function should complete without raising, handling the ParserSyntaxError internally


async def test_libcst_transform_handles_cst_validation_error():
    """Test that libcst_transform handles CSTValidationError gracefully.

    A transformer that raises CSTValidationError should not crash the reduction.
    """

    # Create a transformer that causes validation errors
    def bad_transformer(node):
        # Try to create an invalid CST node by adding a node in wrong context
        # This should trigger CSTValidationError
        raise libcst.CSTValidationError("Test validation error")

    code = b"x = 1\n"

    async def is_interesting(x: bytes) -> bool:
        return True

    problem: BasicReductionProblem[bytes] = BasicReductionProblem(
        initial=code,
        is_interesting=is_interesting,
        work=WorkContext(parallelism=1),
    )
    await libcst_transform(
        problem,
        m.SimpleStatementLine(),
        bad_transformer,
    )
    # Should complete without raising


async def test_libcst_transform_handles_type_error_does_not_allow():
    """Test that libcst_transform handles TypeError with 'does not allow for it'.

    This error occurs when a transformation is structurally invalid for the parent node.
    """

    # Create a transformer that raises the specific TypeError
    def bad_transformer(node):
        raise TypeError("The parent node does not allow for it")

    code = b"x = 1\n"

    async def is_interesting(x: bytes) -> bool:
        return True

    problem: BasicReductionProblem[bytes] = BasicReductionProblem(
        initial=code,
        is_interesting=is_interesting,
        work=WorkContext(parallelism=1),
    )
    await libcst_transform(
        problem,
        m.SimpleStatementLine(),
        bad_transformer,
    )
    # Should complete without raising


async def test_libcst_transform_reraises_other_type_error():
    """Test that libcst_transform re-raises TypeError without 'does not allow for it'.

    TypeErrors that aren't about parent node restrictions should propagate.
    """

    # Create a transformer that raises a different TypeError
    def bad_transformer(node):
        raise TypeError("Some other type error")

    code = b"x = 1\n"

    async def is_interesting(x: bytes) -> bool:
        return True

    problem: BasicReductionProblem[bytes] = BasicReductionProblem(
        initial=code,
        is_interesting=is_interesting,
        work=WorkContext(parallelism=1),
    )
    # The TypeError gets wrapped in ExceptionGroups by trio nurseries
    with pytest.raises(ExceptionGroup):
        await libcst_transform(
            problem,
            m.SimpleStatementLine(),
            bad_transformer,
        )



"""UI abstractions for shrink ray."""

from abc import ABC
from typing import TYPE_CHECKING

import humanize
import trio
from attrs import define


if TYPE_CHECKING:
    from shrinkray.problem import BasicReductionProblem
    from shrinkray.state import ShrinkRayState


@define(slots=False)
class ShrinkRayUI[TestCase](ABC):
    """Base class for shrink ray UI implementations."""

    state: "ShrinkRayState[TestCase]"

    @property
    def reducer(self):
        return self.state.reducer

    @property
    def problem(self) -> "BasicReductionProblem":
        return self.reducer.target  # type: ignore

    def install_into_nursery(self, nursery: trio.Nursery):  # noqa: B027
        """Install any background tasks into the nursery.

        Subclasses may override this to add tasks that run alongside reduction.
        The default implementation does nothing.
        """

    async def run(self, nursery: trio.Nursery):  # noqa: B027
        """Run the UI update loop.

        Subclasses may override this to display progress or status updates.
        The default implementation does nothing.
        """


class BasicUI[TestCase](ShrinkRayUI[TestCase]):
    """Simple text-based UI for non-interactive use."""

    async def run(self, nursery: trio.Nursery):
        initial = self.state.initial
        size = self.state.problem.size
        print(
            f"Starting reduction. Initial test case size: "
            f"{humanize.naturalsize(size(initial))}",
            flush=True,
        )
        prev_reduction = 0
        while True:
            initial = self.state.initial
            current = self.state.problem.current_test_case
            size = self.state.problem.size
            reduction = size(initial) - size(current)
            if reduction > prev_reduction:
                print(
                    f"Reduced test case to {humanize.naturalsize(size(current))} "
                    f"(deleted {humanize.naturalsize(reduction)}, "
                    f"{humanize.naturalsize(reduction - prev_reduction)} since last time)",
                    flush=True,
                )
                prev_reduction = reduction
                await trio.sleep(5)
            else:
                await trio.sleep(0.1)



"""Snapshot tests for the textual TUI."""

import asyncio
from pathlib import Path

import pytest

from shrinkray.subprocess.protocol import PassStatsData, ProgressUpdate, Response
from shrinkray.tui import ShrinkRayApp, StatsDisplay


# Directory where snapshots are stored
SNAPSHOTS_DIR = Path(__file__).parent / "__snapshots__" / "test_tui_snapshots"


def pytest_sessionfinish(session, exitstatus):
    """Verify all snapshots contain expected content after tests complete."""
    if not SNAPSHOTS_DIR.exists():
        return

    # Using vendored pytest_textual_snapshot with syrupy 5.0 compatibility
    for snapshot_file in SNAPSHOTS_DIR.glob("*.svg"):
        content = snapshot_file.read_text()
        # Each snapshot should contain "Reducer" which appears in StatsDisplay
        # Note: SVG uses &#160; for non-breaking spaces, so we just check for "Reducer"
        if "Reducer" not in content:
            raise AssertionError(
                f"Snapshot {snapshot_file.name} does not contain expected text 'Reducer'. "
                "The snapshot may have been captured before the app fully rendered."
            )


class FakeReductionClientForSnapshots:
    """Fake client that provides controlled updates for snapshot testing."""

    def __init__(self, updates: list[ProgressUpdate]):
        self._updates = updates
        self._update_index = 0
        self._cancelled = False
        self._updates_consumed = False

    async def start(self) -> None:
        pass

    async def start_reduction(
        self,
        file_path: str,
        test: list[str],
        parallelism: int | None = None,
        timeout: float | None = None,
        seed: int = 0,
        input_type: str = "all",
        in_place: bool = False,
        formatter: str = "default",
        volume: str = "normal",
        no_clang_delta: bool = False,
        clang_delta: str = "",
        trivial_is_error: bool = True,
    ) -> Response:
        return Response(id="start", result={"status": "started"})

    async def cancel(self) -> Response:
        self._cancelled = True
        return Response(id="cancel", result={"status": "cancelled"})

    async def disable_pass(self, pass_name: str) -> Response:
        return Response(
            id="disable", result={"status": "disabled", "pass_name": pass_name}
        )

    async def enable_pass(self, pass_name: str) -> Response:
        return Response(
            id="enable", result={"status": "enabled", "pass_name": pass_name}
        )

    async def skip_current_pass(self) -> Response:
        return Response(id="skip", result={"status": "skipped"})

    async def close(self) -> None:
        pass

    async def get_progress_updates(self):
        for update in self._updates:
            if self._cancelled:
                break
            yield update
        self._updates_consumed = True
        # Keep the app running by waiting indefinitely (until cancelled)
        while not self._cancelled:
            await asyncio.sleep(1)

    @property
    def is_completed(self) -> bool:
        # Only complete when cancelled, not when updates are consumed
        return self._cancelled

    @property
    def error_message(self) -> str | None:
        return None


def make_app_with_updates(updates: list[ProgressUpdate]) -> ShrinkRayApp:
    """Create a ShrinkRayApp with fake updates for testing."""
    client = FakeReductionClientForSnapshots(updates)
    return ShrinkRayApp(
        file_path="/tmp/test.txt",
        test=["./test.sh"],
        client=client,
        theme="dark",  # Explicit theme for deterministic snapshots
    )


@pytest.fixture
def initial_state_update() -> ProgressUpdate:
    """A progress update showing initial state."""
    return ProgressUpdate(
        status="Starting reduction",
        size=10000,
        original_size=10000,
        calls=0,
        reductions=0,
        interesting_calls=0,
        wasted_calls=0,
        runtime=0.0,
        parallel_workers=4,
        average_parallelism=4.0,
        effective_parallelism=4.0,
        time_since_last_reduction=0.0,
        content_preview="Line 1\nLine 2\nLine 3\nLine 4\nLine 5",
        hex_mode=False,
    )


@pytest.fixture
def mid_reduction_update() -> ProgressUpdate:
    """A progress update showing mid-reduction state."""
    return ProgressUpdate(
        status="Running DeleteChunks",
        size=5000,
        original_size=10000,
        calls=150,
        reductions=25,
        interesting_calls=50,
        wasted_calls=10,
        runtime=30.5,
        parallel_workers=4,
        average_parallelism=3.8,
        effective_parallelism=3.5,
        time_since_last_reduction=2.5,
        content_preview="Line 1\nLine 3\nLine 5",
        hex_mode=False,
    )


@pytest.fixture
def hex_mode_update() -> ProgressUpdate:
    """A progress update showing hex mode."""
    return ProgressUpdate(
        status="Running ByteReduction",
        size=256,
        original_size=512,
        calls=50,
        reductions=10,
        interesting_calls=20,
        wasted_calls=5,
        runtime=10.0,
        parallel_workers=2,
        average_parallelism=2.0,
        effective_parallelism=1.8,
        time_since_last_reduction=1.0,
        content_preview="00000000  48 65 6c 6c 6f 20 57 6f  72 6c 64 21              Hello World!",
        hex_mode=True,
    )


@pytest.fixture
def large_file_update() -> ProgressUpdate:
    """A progress update with large file content for diff testing."""
    # Create content with many lines
    lines = [f"Line {i}: some content here" for i in range(100)]
    return ProgressUpdate(
        status="Running pass 5",
        size=50000,
        original_size=100000,
        calls=500,
        reductions=100,
        interesting_calls=200,
        wasted_calls=50,
        runtime=120.0,
        parallel_workers=8,
        average_parallelism=7.5,
        effective_parallelism=6.8,
        time_since_last_reduction=0.5,
        content_preview="\n".join(lines),
        hex_mode=False,
    )


# === TUI snapshot tests ===


async def wait_for_render(pilot):
    """Wait for the app to fully render with actual content.

    Checks that the StatsDisplay has received and rendered the update
    by verifying original_size is non-zero (all test fixtures have non-zero size).
    """
    max_attempts = 50  # Generous limit to avoid flakiness
    for _ in range(max_attempts):
        await pilot.pause()
        stats = pilot.app.query_one("#stats-display", StatsDisplay)
        if stats.original_size > 0:
            # Content has been rendered, give one more pause for layout
            await pilot.pause()
            return

    # Get final state for error message
    final_stats = pilot.app.query_one("#stats-display", StatsDisplay)
    raise AssertionError(
        "App did not render within expected time. "
        f"StatsDisplay.original_size is still {final_stats.original_size}"
    )


def test_initial_state(snap_compare, initial_state_update):
    """Snapshot test for initial app state."""
    app = make_app_with_updates([initial_state_update])
    assert snap_compare(app, terminal_size=(120, 40), run_before=wait_for_render)


def test_mid_reduction(snap_compare, mid_reduction_update):
    """Snapshot test for mid-reduction state."""
    app = make_app_with_updates([mid_reduction_update])
    assert snap_compare(app, terminal_size=(120, 40), run_before=wait_for_render)


def test_hex_mode(snap_compare, hex_mode_update):
    """Snapshot test for hex mode display."""
    app = make_app_with_updates([hex_mode_update])
    assert snap_compare(app, terminal_size=(120, 40), run_before=wait_for_render)


def test_large_file(snap_compare, large_file_update):
    """Snapshot test for large file with potential diff."""
    app = make_app_with_updates([large_file_update])
    assert snap_compare(app, terminal_size=(120, 40), run_before=wait_for_render)


def test_small_terminal(snap_compare, mid_reduction_update):
    """Snapshot test with smaller terminal size."""
    app = make_app_with_updates([mid_reduction_update])
    assert snap_compare(app, terminal_size=(80, 24), run_before=wait_for_render)


def test_wide_terminal(snap_compare, mid_reduction_update):
    """Snapshot test with wider terminal."""
    app = make_app_with_updates([mid_reduction_update])
    assert snap_compare(app, terminal_size=(160, 50), run_before=wait_for_render)


@pytest.fixture
def update_with_pass_stats() -> ProgressUpdate:
    """A progress update with pass statistics."""
    return ProgressUpdate(
        status="Running hollow",
        size=5000,
        original_size=10000,
        calls=150,
        reductions=25,
        interesting_calls=50,
        wasted_calls=10,
        runtime=30.5,
        parallel_workers=4,
        average_parallelism=3.8,
        effective_parallelism=3.5,
        time_since_last_reduction=2.5,
        content_preview="Line 1\nLine 3\nLine 5",
        hex_mode=False,
        pass_stats=[
            PassStatsData(
                pass_name="hollow",
                bytes_deleted=2500,
                run_count=5,
                test_evaluations=100,
                successful_reductions=3,
                success_rate=3.0,
            ),
            PassStatsData(
                pass_name="delete_lines",
                bytes_deleted=1500,
                run_count=3,
                test_evaluations=80,
                successful_reductions=2,
                success_rate=2.5,
            ),
            PassStatsData(
                pass_name="lift_braces",
                bytes_deleted=1000,
                run_count=2,
                test_evaluations=50,
                successful_reductions=1,
                success_rate=2.0,
            ),
        ],
        current_pass_name="hollow",
        disabled_passes=[],
    )


async def test_pass_stats_modal(snap_compare, update_with_pass_stats):
    """Snapshot test for the pass statistics modal."""

    async def interact(pilot):
        # Wait a moment for the app to start
        await pilot.pause()
        # Press 'p' to open the pass stats modal
        await pilot.press("p")
        # Wait for the modal to render
        await pilot.pause()

    app = make_app_with_updates([update_with_pass_stats])
    assert snap_compare(app, terminal_size=(120, 40), run_before=interact)



"""CLI utilities and types for shrink ray."""

import os
import shlex
import sys
from enum import Enum, IntEnum, auto
from shutil import which
from typing import Any

import click


def validate_command(ctx: Any, param: Any, value: str) -> list[str]:
    """Validate and resolve a command string."""
    parts = shlex.split(value)
    command = parts[0]

    if os.path.exists(command):
        command = os.path.abspath(command)
    else:
        what = which(command)
        if what is None:
            raise click.BadParameter(f"{command}: command not found")
        command = os.path.abspath(what)
    return [command] + parts[1:]


class EnumChoice[EnumType: Enum](click.Choice):
    """A click Choice that works with Enums."""

    def __init__(self, enum: type[EnumType]) -> None:
        self.enum = enum
        choices = [str(e.name) for e in enum]
        self.__values = {e.name: e for e in enum}
        super().__init__(choices)

    def convert(self, value: str, param: Any, ctx: Any) -> EnumType:
        return self.__values[value]


class InputType(IntEnum):
    """How input is passed to the test function."""

    all = 0
    stdin = 1
    arg = 2
    basename = 3

    def enabled(self, value: "InputType") -> bool:
        if self == InputType.all:
            return True
        return self == value


class UIType(Enum):
    """Type of UI to use."""

    basic = auto()
    textual = auto()


def validate_ui(ctx, param, value) -> UIType:
    """Validate and determine UI type."""
    if value is None:
        if sys.stdin.isatty() and sys.stdout.isatty():
            return UIType.textual
        else:
            return UIType.basic
    else:
        return value



"""Main entry point for shrink ray."""

import os
import shutil
import signal
import sys
import traceback
from typing import Any

import click
import trio

from shrinkray.cli import (
    EnumChoice,
    InputType,
    UIType,
    validate_command,
    validate_ui,
)
from shrinkray.passes.clangdelta import (
    C_FILE_EXTENSIONS,
    ClangDelta,
    find_clang_delta,
)
from shrinkray.problem import InvalidInitialExample
from shrinkray.state import (
    ShrinkRayDirectoryState,
    ShrinkRayState,
    ShrinkRayStateSingleFile,
)
from shrinkray.ui import BasicUI, ShrinkRayUI
from shrinkray.work import Volume


async def run_shrink_ray(
    state: ShrinkRayState[Any],
    ui: ShrinkRayUI[Any],
) -> None:
    """Run the shrink ray reduction process."""
    async with trio.open_nursery() as nursery:
        problem = state.problem
        try:
            await problem.setup()
        except* InvalidInitialExample as excs:
            assert len(excs.exceptions) == 1
            (e,) = excs.exceptions
            await state.report_error(e)

        reducer = state.reducer

        @nursery.start_soon
        async def _() -> None:
            await reducer.run()
            nursery.cancel_scope.cancel()

        ui.install_into_nursery(nursery)

        await ui.run(nursery)

    await state.print_exit_message(problem)


@click.command(
    help="""
""".strip()
)
@click.version_option()
@click.option(
    "--backup",
    default="",
    help=(
        "Name of the backup file to create. Defaults to adding .bak to the "
        "name of the source file"
    ),
)
@click.option(
    "--timeout",
    default=None,
    type=click.FLOAT,
    help=(
        "Time out subprocesses after this many seconds. If not specified, "
        "runs the interestingness test once and sets timeout to 10x the "
        "measured time (capped at 5 minutes). If set to <= 0 then no timeout "
        "will be used. Any commands that time out will be treated as failing "
        "the test"
    ),
)
@click.option(
    "--seed",
    default=0,
    type=click.INT,
    help=("Random seed to use for any non-deterministic reductions."),
)
@click.option(
    "--volume",
    default="normal",
    type=EnumChoice(Volume),
    help="Level of output to provide.",
)
@click.option(
    "--in-place/--not-in-place",
    default=False,
    help="""
If `--in-place` is passed, shrinkray will run in the current working directory instead of
creating a temporary subdirectory. Note that this requires you to either run with no
parallelism or be very careful about files created in your interestingness
test not conflicting with each other.
""",
)
@click.option(
    "--input-type",
    default="all",
    type=EnumChoice(InputType),
    help="""
How to pass input to the test function. Options are:

1. `basename` writes it to a file of the same basename as the original, in the current working directory where the test is run.

2. `arg` passes it in a file whose name is provided as an argument to the test.

3. `stdin` passes its contents on stdin.

4. `all` (the default) does all of the above.

If --in-place is specified, all will not include basename by default, only arg and stdin.
If you want basename with --in-place you may pass it explicitly, but note that this is incompatible
with any parallelism.
    """.strip(),
)
@click.option(
    "--parallelism",
    type=click.INT,
    help="Number of tests to run in parallel. If set to 0 will default to either 1 or number of cpus depending on other options.",
    default=0,
)
@click.option(
    "--ui",
    "ui_type",
    type=EnumChoice(UIType),
    help="""
UI mode to use. Options are:

* 'textual' (default): Modern terminal UI using the textual library.
* 'basic': Simple text output, suitable for scripts or non-interactive use.

When not specified, defaults to 'textual' for interactive terminals, 'basic' otherwise.
    """.strip(),
    callback=validate_ui,
)
@click.option(
    "--theme",
    type=click.Choice(["auto", "dark", "light"]),
    default="auto",
    help="""
Theme mode for the textual UI. Options are:

* 'auto' (default): Detect terminal's color scheme automatically.
* 'dark': Use dark theme.
* 'light': Use light theme.
    """.strip(),
)
@click.option(
    "--formatter",
    default="default",
    help="""
Path to a formatter for Shrink Ray to use. This is mostly used for display purposes,
and to format the final test case.

A formatter should accept input on stdin and write to stdout, and exit with a status
code of 0. If the formatter exits with a non-zero status code its output will be
ignored.

Special values for this:

* 'none' turns off formatting.
* 'default' causes Shrink Ray to use its default behaviour, which is to look for
  formatters it knows about on PATH and use one of those if found, otherwise to
  use a very simple language-agnostic formatter.
""",
)
@click.option(
    "--trivial-is-error/--trivial-is-not-error",
    default=True,
    help="""
It's easy to write interestingness tests which accept too much, and one common way this
happens is if they accept empty or otherwise trivial files. By default Shrink Ray will
print an error message at the end of reduction and exit with non-zero status in this case.
This behaviour can be disabled by passing --trivial-is-not-error.
""",
)
@click.option(
    "--exit-on-completion/--no-exit-on-completion",
    default=True,
    help="Exit automatically when reduction completes (TUI only). Default: exit on completion.",
)
@click.option(
    "--no-clang-delta",
    is_flag=True,
    default=False,
    help="Pass this if you do not want to use clang delta for C/C++ transformations.",
)
@click.option(
    "--clang-delta",
    default="",
    help="Path to your clang_delta executable.",
)
@click.argument("test", callback=validate_command)
@click.argument(
    "filename",
    type=click.Path(exists=True, resolve_path=False, dir_okay=True, allow_dash=False),
)
def main(
    input_type: InputType,
    backup: str,
    filename: str,
    test: list[str],
    timeout: float | None,
    in_place: bool,
    parallelism: int,
    seed: int,
    volume: Volume,
    formatter: str,
    no_clang_delta: bool,
    clang_delta: str,
    trivial_is_error: bool,
    exit_on_completion: bool,
    ui_type: UIType,
    theme: str,
) -> None:
    if timeout is not None and timeout <= 0:
        timeout = float("inf")

    if not os.access(test[0], os.X_OK):
        print(
            f"Interestingness test {os.path.relpath(test[0])} is not executable.",
            file=sys.stderr,
        )
        sys.exit(1)

    if in_place and input_type == InputType.basename and parallelism > 1:
        raise click.BadParameter(
            f"parallelism cannot be greater than 1 when --in-place and --input-type=basename (got {parallelism})"
        )

    if parallelism == 0:
        if in_place and input_type == InputType.basename:
            parallelism = 1
        else:
            parallelism = os.cpu_count() or 1

    clang_delta_executable: ClangDelta | None = None
    if os.path.splitext(filename)[1] in C_FILE_EXTENSIONS and not no_clang_delta:
        if not clang_delta:
            clang_delta = find_clang_delta()
        if not clang_delta:
            raise click.UsageError(
                "Attempting to reduce a C or C++ file, but clang_delta is not installed. "
                "Please run with --no-clang-delta, or install creduce on your system. "
                "If creduce is already installed and you wish to use clang_delta, please "
                "pass its path with the --clang-delta argument."
            )

        clang_delta_executable = ClangDelta(clang_delta)

    # This is a debugging option so that when the reducer seems to be taking
    # a long time you can Ctrl-\ to find out what it's up to. I have no idea
    # how to test it in a way that shows up in coverage.
    def dump_trace(signum: int, frame: Any) -> None:  # pragma: no cover
        traceback.print_stack()

    signal.signal(signal.SIGQUIT, dump_trace)

    if not backup:
        backup = filename + os.extsep + "bak"

    state_kwargs: dict[str, Any] = {
        "input_type": input_type,
        "in_place": in_place,
        "test": test,
        "timeout": timeout,
        "base": os.path.basename(filename),
        "parallelism": parallelism,
        "filename": filename,
        "formatter": formatter,
        "trivial_is_error": trivial_is_error,
        "seed": seed,
        "volume": volume,
        "clang_delta_executable": clang_delta_executable,
    }

    state: ShrinkRayState[Any]
    ui: ShrinkRayUI[Any]

    if os.path.isdir(filename):
        if input_type == InputType.stdin:
            raise click.UsageError("Cannot pass a directory input on stdin.")

        shutil.rmtree(backup, ignore_errors=True)
        shutil.copytree(filename, backup)

        files = [os.path.join(d, f) for d, _, fs in os.walk(filename) for f in fs]

        initial = {}
        for f in files:
            with open(f, "rb") as i:
                initial[os.path.relpath(f, filename)] = i.read()

        state = ShrinkRayDirectoryState(initial=initial, **state_kwargs)

        trio.run(state.check_formatter)

    else:
        try:
            os.remove(backup)
        except FileNotFoundError:
            pass

        with open(filename, "rb") as reader:
            initial = reader.read()

        with open(backup, "wb") as writer:
            writer.write(initial)

        state = ShrinkRayStateSingleFile(initial=initial, **state_kwargs)

        trio.run(state.check_formatter)

    if ui_type == UIType.textual:
        from shrinkray.tui import run_textual_ui

        run_textual_ui(
            file_path=filename,
            test=test,
            parallelism=parallelism,
            timeout=timeout,
            seed=seed,
            input_type=input_type.name,
            in_place=in_place,
            formatter=formatter,
            volume=volume.name,
            no_clang_delta=no_clang_delta,
            clang_delta=clang_delta,
            trivial_is_error=trivial_is_error,
            exit_on_completion=exit_on_completion,
            theme=theme,  # type: ignore[arg-type]
        )
        return

    # At this point, ui_type must be UIType.basic since textual returned above
    assert ui_type == UIType.basic
    ui = BasicUI(state)

    try:
        trio.run(
            lambda: run_shrink_ray(
                state=state,
                ui=ui,
            )
        )
    # If you try to sys.exit from within an exception handler, trio will instead
    # put it in an exception group. I wish to register the complaint that this is
    # incredibly fucking stupid, but anyway this is a workaround for it.
    except* SystemExit as eg:
        raise eg.exceptions[0]
    except* KeyboardInterrupt as eg:
        raise eg.exceptions[0]


def worker_main() -> None:
    """Entry point for the worker subprocess."""
    from shrinkray.subprocess.worker import main as worker_entry

    worker_entry()


if __name__ == "__main__":  # pragma: no cover
    main(prog_name="shrinkray")



"""Tests for process management utilities."""

import os
import signal
import subprocess
import sys
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
import trio

from shrinkray.process import interrupt_wait_and_kill, signal_group


# === signal_group tests ===


async def test_signal_group_sends_signal_to_process_group():
    # Start a process in its own process group
    sp = await trio.lowlevel.open_process(
        [sys.executable, "-c", "import time; time.sleep(100)"],
        preexec_fn=os.setsid,
    )
    try:
        # Process should be running
        assert sp.poll() is None

        # Send SIGTERM to the group
        signal_group(sp, signal.SIGTERM)

        # Wait for process to exit
        with trio.move_on_after(5):
            await sp.wait()

        assert sp.returncode is not None
    finally:
        if sp.returncode is None:
            sp.kill()
            await sp.wait()


# === interrupt_wait_and_kill tests ===


async def test_interrupt_wait_and_kill_does_nothing_if_already_exited():
    # Start a process that exits immediately
    sp = await trio.lowlevel.open_process(
        [sys.executable, "-c", "pass"],
        preexec_fn=os.setsid,
    )
    # Wait for it to exit
    await sp.wait()
    assert sp.returncode is not None

    # Should not raise
    await interrupt_wait_and_kill(sp)


@pytest.mark.slow
async def test_interrupt_wait_and_kill_kills_process_ignoring_sigint():
    # Start a process that ignores SIGINT
    sp = await trio.lowlevel.open_process(
        [
            sys.executable,
            "-c",
            "import signal, time; signal.signal(signal.SIGINT, lambda *a: None); time.sleep(100)",
        ],
        preexec_fn=os.setsid,
        stdout=subprocess.PIPE,
    )

    # Give it a moment to set up the signal handler
    await trio.sleep(0.1)

    # This should eventually kill it
    await interrupt_wait_and_kill(sp, delay=0.05)

    assert sp.returncode is not None


async def test_interrupt_wait_and_kill_handles_fast_exit_after_sigint():
    # Start a process that exits quickly after SIGINT
    sp = await trio.lowlevel.open_process(
        [
            sys.executable,
            "-c",
            "import signal, time; signal.signal(signal.SIGINT, lambda *a: exit(0)); time.sleep(100)",
        ],
        preexec_fn=os.setsid,
    )

    await trio.sleep(0.05)
    await interrupt_wait_and_kill(sp, delay=0.05)

    assert sp.returncode is not None
    assert sp.returncode == 0


async def test_interrupt_wait_and_kill_closes_pipes_before_signaling():
    # Start a process with stdout pipe
    sp = await trio.lowlevel.open_process(
        [
            sys.executable,
            "-c",
            "import sys; print('hello'); sys.stdout.flush(); import time; time.sleep(100)",
        ],
        preexec_fn=os.setsid,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )

    # Give it a moment to print
    await trio.sleep(0.1)

    # Should close pipes and kill
    await interrupt_wait_and_kill(sp, delay=0.05)

    assert sp.returncode is not None


async def test_interrupt_wait_and_kill_handles_process_lookup_error_on_sigkill():
    """Test that ProcessLookupError is handled when sending SIGKILL.

    The process exits between checking returncode and sending SIGKILL.
    """
    # Create a mock process
    mock_sp = MagicMock()
    mock_sp.pid = 12345
    mock_sp.stdout = None
    mock_sp.stderr = None
    mock_sp.stdin = None

    # returncode is None initially, then None at returncode check,
    # but process disappears before SIGKILL
    poll_returns = iter([None] * 11)  # All poll() calls return None
    mock_sp.poll.side_effect = lambda: next(poll_returns)

    # Track when SIGKILL is attempted
    sigkill_attempted = [False]

    # returncode property: None initially, then 0 after SIGKILL is attempted
    # (process exits between the check and the kill)
    def get_returncode():
        if sigkill_attempted[0]:
            return 0  # Process has exited
        return None

    type(mock_sp).returncode = property(lambda self: get_returncode())

    mock_sp.wait = AsyncMock()

    # signal_group should work for SIGINT but raise ProcessLookupError for SIGKILL
    def mock_signal_group(sp, sig):
        if sig == signal.SIGKILL:
            sigkill_attempted[0] = True
            raise ProcessLookupError("No such process")
        # Don't actually send signal to avoid affecting real processes

    with patch("shrinkray.process.signal_group", mock_signal_group):
        # Should not raise despite ProcessLookupError
        await interrupt_wait_and_kill(mock_sp, delay=0.001)


async def test_interrupt_wait_and_kill_raises_on_unkillable_process():
    """Test that ValueError is raised when process cannot be killed.

    The process persists after all kill attempts.
    """
    # Create a mock process that never dies
    mock_sp = MagicMock()
    mock_sp.pid = 12345
    mock_sp.stdout = None
    mock_sp.stderr = None
    mock_sp.stdin = None

    # Always return None for poll (process never exits from SIGINT)
    mock_sp.poll.return_value = None

    # returncode is always None (process never dies)
    type(mock_sp).returncode = property(lambda self: None)

    mock_sp.wait = AsyncMock()

    # signal_group does nothing (signal is ignored)
    with patch("shrinkray.process.signal_group"):
        with pytest.raises(ValueError, match="Could not kill subprocess"):
            await interrupt_wait_and_kill(mock_sp, delay=0.001)


async def test_interrupt_wait_and_kill_skips_sigkill_if_process_exits_after_poll_loop():
    """Test that SIGKILL is skipped if process exits after poll loop completes.

    The returncode becomes non-None after all poll() calls return None but
    before checking returncode for the SIGKILL decision.
    """
    # Create a mock process
    mock_sp = MagicMock()
    mock_sp.pid = 12345
    mock_sp.stdout = None
    mock_sp.stderr = None
    mock_sp.stdin = None

    # poll() always returns None during the loop (so loop completes all iterations)
    mock_sp.poll.return_value = None

    # Track when we're past the poll loop
    poll_loop_done = [False]

    # returncode: None during poll loop, 0 after (simulating process exiting
    # after the loop but before the returncode check)
    call_count = [0]

    def counting_poll():
        call_count[0] += 1
        if call_count[0] >= 10:
            poll_loop_done[0] = True
        return None

    mock_sp.poll.side_effect = counting_poll

    def get_returncode():
        if poll_loop_done[0]:
            return 0  # Process exited after the poll loop
        return None

    type(mock_sp).returncode = property(lambda self: get_returncode())

    mock_sp.wait = AsyncMock()

    sigkill_sent = [False]

    def mock_signal_group(sp, sig):
        if sig == signal.SIGKILL:
            sigkill_sent[0] = True

    with patch("shrinkray.process.signal_group", mock_signal_group):
        await interrupt_wait_and_kill(mock_sp, delay=0.001)

    # SIGKILL should not have been sent since returncode became non-None
    # after the poll loop, skipping the SIGKILL branch
    assert not sigkill_sent[0]



"""State management for shrink ray reduction sessions."""

import math
import os
import random
import shutil
import subprocess
import sys
import time
from abc import ABC, abstractmethod
from datetime import timedelta
from tempfile import TemporaryDirectory
from typing import Any

import humanize
import trio
from attrs import define

from shrinkray.passes.clangdelta import ClangDelta
from shrinkray.problem import (
    BasicReductionProblem,
    InvalidInitialExample,
    ReductionProblem,
    shortlex,
)
from shrinkray.reducer import DirectoryShrinkRay, Reducer, ShrinkRay
from shrinkray.work import Volume, WorkContext


class TimeoutExceededOnInitial(InvalidInitialExample):
    def __init__(self, runtime: float, timeout: float) -> None:
        self.runtime = runtime
        self.timeout = timeout
        super().__init__(
            f"Initial test call exceeded timeout of {timeout}s. Try raising or disabling timeout."
        )


# Constants for dynamic timeout
DYNAMIC_TIMEOUT_CALIBRATION_TIMEOUT = 300.0  # 5 minutes for first call
DYNAMIC_TIMEOUT_MULTIPLIER = 10
DYNAMIC_TIMEOUT_MAX = 300.0  # 5 minutes maximum
DYNAMIC_TIMEOUT_MIN = 1.0  # 1 second minimum to prevent edge cases


def compute_dynamic_timeout(runtime: float) -> float:
    """Compute dynamic timeout based on measured runtime.

    The timeout is set to 10x the measured runtime, clamped between
    DYNAMIC_TIMEOUT_MIN and DYNAMIC_TIMEOUT_MAX.
    """
    return max(
        DYNAMIC_TIMEOUT_MIN,
        min(runtime * DYNAMIC_TIMEOUT_MULTIPLIER, DYNAMIC_TIMEOUT_MAX),
    )


@define(slots=False)
class ShrinkRayState[TestCase](ABC):
    input_type: Any  # InputType from __main__
    in_place: bool
    test: list[str]
    filename: str
    timeout: float | None
    base: str
    parallelism: int
    initial: TestCase
    formatter: str
    trivial_is_error: bool
    seed: int
    volume: Volume
    clang_delta_executable: ClangDelta | None

    first_call: bool = True
    initial_exit_code: int | None = None
    can_format: bool = True
    formatter_command: list[str] | None = None

    first_call_time: float | None = None

    # Lazy imports to break circular dependencies:
    # - shrinkray.process imports from shrinkray.work which imports from here
    # - shrinkray.cli imports from here for state configuration
    # These are cached after first import for performance.
    _interrupt_wait_and_kill: Any = None
    _InputType: Any = None  # InputType enum from shrinkray.cli

    # Stores the output from the last debug run
    _last_debug_output: str = ""

    def __attrs_post_init__(self):
        self.is_interesting_limiter = trio.CapacityLimiter(max(self.parallelism, 1))
        self.setup_formatter()

    @abstractmethod
    def setup_formatter(self): ...

    @abstractmethod
    def new_reducer(self, problem: ReductionProblem[TestCase]) -> Reducer[TestCase]: ...

    @abstractmethod
    async def write_test_case_to_file_impl(self, working: str, test_case: TestCase): ...

    async def write_test_case_to_file(self, working: str, test_case: TestCase):
        await self.write_test_case_to_file_impl(working, test_case)

    async def run_script_on_file(
        self, working: str, cwd: str, debug: bool = False
    ) -> int:
        # Lazy import to avoid circular dependency
        if self._interrupt_wait_and_kill is None:
            from shrinkray.process import interrupt_wait_and_kill

            self._interrupt_wait_and_kill = interrupt_wait_and_kill
        if self._InputType is None:
            from shrinkray.cli import InputType

            self._InputType = InputType

        if not os.path.exists(working):
            raise ValueError(f"No such file {working}")
        if self.input_type.enabled(self._InputType.arg):
            command = self.test + [working]
        else:
            command = self.test

        kwargs: dict[str, Any] = {
            "universal_newlines": False,
            "preexec_fn": os.setsid,
            "cwd": cwd,
            "check": False,
        }
        if self.input_type.enabled(self._InputType.stdin) and not os.path.isdir(
            working
        ):
            with open(working, "rb") as i:
                kwargs["stdin"] = i.read()
        else:
            kwargs["stdin"] = b""

        # For debug mode, use simpler approach to capture output
        if debug:
            kwargs["capture_stdout"] = True
            kwargs["capture_stderr"] = True
            start_time = time.time()
            completed = await trio.run_process(command, **kwargs)
            runtime = time.time() - start_time

            # Check for timeout violation (only when timeout is explicitly set)
            if self.timeout is not None and runtime >= self.timeout and self.first_call:
                self.initial_exit_code = completed.returncode
                self.first_call = False
                raise TimeoutExceededOnInitial(
                    timeout=self.timeout,
                    runtime=runtime,
                )

            if self.first_call:
                self.initial_exit_code = completed.returncode
                # Set dynamic timeout if not explicitly specified
                if self.timeout is None:
                    self.timeout = compute_dynamic_timeout(runtime)
            self.first_call = False

            # Store captured output
            output_parts = []
            if completed.stdout:
                output_parts.append(completed.stdout.decode("utf-8", errors="replace"))
            if completed.stderr:
                output_parts.append(completed.stderr.decode("utf-8", errors="replace"))
            self._last_debug_output = "\n".join(output_parts).strip()

            return completed.returncode

        # Check if we should stream output to stderr (volume=debug)
        if self.volume == Volume.debug:
            # Inherit stderr from parent process to stream output in real-time
            kwargs["stderr"] = None  # None means inherit
            kwargs["stdout"] = subprocess.DEVNULL
        else:
            # Non-debug mode: discard all output
            kwargs["stdout"] = subprocess.DEVNULL
            kwargs["stderr"] = subprocess.DEVNULL

        async with trio.open_nursery() as nursery:

            def call_with_kwargs(task_status=trio.TASK_STATUS_IGNORED):  # type: ignore
                return trio.run_process(command, **kwargs, task_status=task_status)

            start_time = time.time()
            sp = await nursery.start(call_with_kwargs)

            try:
                # Determine effective timeout for this call
                if self.first_call:
                    # For first call: use calibration timeout if dynamic, otherwise 10x explicit timeout
                    if self.timeout is None:
                        effective_timeout = DYNAMIC_TIMEOUT_CALIBRATION_TIMEOUT
                    else:
                        effective_timeout = self.timeout * 10
                else:
                    # For subsequent calls, timeout must be set (either explicit or computed)
                    assert self.timeout is not None
                    effective_timeout = self.timeout

                with trio.move_on_after(effective_timeout):
                    await sp.wait()

                runtime = time.time() - start_time

                if sp.returncode is None:
                    # Process didn't terminate before timeout - kill it
                    await self._interrupt_wait_and_kill(sp)

                # Check for timeout violation (only when timeout is explicitly set)
                if (
                    self.timeout is not None
                    and runtime >= self.timeout
                    and self.first_call
                ):
                    raise TimeoutExceededOnInitial(
                        timeout=self.timeout,
                        runtime=runtime,
                    )
            finally:
                if self.first_call:
                    self.initial_exit_code = sp.returncode
                    # Set dynamic timeout if not explicitly specified
                    if self.timeout is None:
                        runtime = time.time() - start_time
                        self.timeout = compute_dynamic_timeout(runtime)
                self.first_call = False

            result: int | None = sp.returncode
            assert result is not None

            return result

    async def run_for_exit_code(self, test_case: TestCase, debug: bool = False) -> int:
        # Lazy import
        if self._InputType is None:
            from shrinkray.cli import InputType

            self._InputType = InputType

        if self.in_place:
            if self.input_type == self._InputType.basename:
                working = self.filename
                await self.write_test_case_to_file(working, test_case)

                return await self.run_script_on_file(
                    working=working,
                    debug=debug,
                    cwd=os.getcwd(),
                )
            else:
                base, ext = os.path.splitext(self.filename)
                working = base + "-" + os.urandom(16).hex() + ext
                assert not os.path.exists(working)
                try:
                    await self.write_test_case_to_file(working, test_case)

                    return await self.run_script_on_file(
                        working=working,
                        debug=debug,
                        cwd=os.getcwd(),
                    )
                finally:
                    if os.path.exists(working):
                        if os.path.isdir(working):
                            shutil.rmtree(working)
                        else:
                            os.unlink(working)
        else:
            with TemporaryDirectory() as d:
                working = os.path.join(d, self.base)
                await self.write_test_case_to_file(working, test_case)

                return await self.run_script_on_file(
                    working=working,
                    debug=debug,
                    cwd=d,
                )

    @abstractmethod
    async def format_data(self, test_case: TestCase) -> TestCase | None: ...

    @abstractmethod
    async def run_formatter_command(
        self, command: str | list[str], input: TestCase
    ) -> subprocess.CompletedProcess: ...

    @abstractmethod
    async def print_exit_message(self, problem): ...

    @property
    def reducer(self):
        try:
            return self.__reducer
        except AttributeError:
            pass

        work = WorkContext(
            random=random.Random(self.seed),
            volume=self.volume,
            parallelism=self.parallelism,
        )

        problem: BasicReductionProblem[TestCase] = BasicReductionProblem(
            is_interesting=self.is_interesting,
            initial=self.initial,
            work=work,
            **self.extra_problem_kwargs,
        )

        # Writing the file back can't be guaranteed atomic, so we put a lock around
        # writing successful reductions back to the original file so we don't
        # write some confused combination of reductions.
        write_lock = trio.Lock()

        @problem.on_reduce
        async def _(test_case: TestCase):
            async with write_lock:
                await self.write_test_case_to_file(self.filename, test_case)

        self.__reducer = self.new_reducer(problem)
        return self.__reducer

    @property
    def extra_problem_kwargs(self):
        return {}

    @property
    def problem(self):
        return self.reducer.target

    async def is_interesting(self, test_case: TestCase) -> bool:
        if self.first_call_time is None:
            self.first_call_time = time.time()
        async with self.is_interesting_limiter:
            return await self.run_for_exit_code(test_case) == 0

    @property
    def parallel_tasks_running(self) -> int:
        """Number of parallel tasks currently running."""
        return self.is_interesting_limiter.borrowed_tokens

    async def attempt_format(self, data: TestCase) -> TestCase:
        if not self.can_format:
            return data
        attempt = await self.format_data(data)
        if attempt is None:
            self.can_format = False
            return data
        if attempt == data or await self.is_interesting(attempt):
            return attempt
        else:
            self.can_format = False
            return data

    async def check_formatter(self):
        if self.formatter_command is None:
            return
        formatter_result = await self.run_formatter_command(
            self.formatter_command, self.initial
        )

        if formatter_result.returncode != 0:
            print(
                "Formatter exited unexpectedly on initial test case. If this is expected, please run with --formatter=none.",
                file=sys.stderr,
            )
            print(
                formatter_result.stderr.decode("utf-8").strip(),
                file=sys.stderr,
            )
            sys.exit(1)
        reformatted = formatter_result.stdout
        if not await self.is_interesting(reformatted) and await self.is_interesting(
            self.initial
        ):
            print(
                "Formatting initial test case made it uninteresting. If this is expected, please run with --formatter=none.",
                file=sys.stderr,
            )
            print(
                formatter_result.stderr.decode("utf-8").strip(),
                file=sys.stderr,
            )
            sys.exit(1)

    async def build_error_message(self, e: Exception) -> str:
        """Build a detailed error message for an invalid initial example.

        This is used by the subprocess worker to provide helpful error messages
        without printing directly to stderr or calling sys.exit.
        """
        lines = [
            "Shrink ray cannot proceed because the initial call of the "
            "interestingness test resulted in an uninteresting test case."
        ]

        if isinstance(e, TimeoutExceededOnInitial):
            lines.append(
                f"This is because your initial test case took {e.runtime:.2f}s "
                f"exceeding your timeout setting of {self.timeout}."
            )
            lines.append(f"Try rerunning with --timeout={math.ceil(e.runtime * 2)}.")
        else:
            lines.append("Rerunning the interestingness test for debugging purposes...")
            exit_code = await self.run_for_exit_code(self.initial, debug=True)
            if exit_code != 0:
                lines.append(
                    f"This exited with code {exit_code}, but the script should "
                    "return 0 for interesting test cases."
                )
                # Include the captured output from the debug run
                if self._last_debug_output:
                    lines.append("\nOutput from the interestingness test:")
                    lines.append(self._last_debug_output)
                local_exit_code = await self.run_script_on_file(
                    working=self.filename,
                    debug=False,
                    cwd=os.getcwd(),
                )
                if local_exit_code == 0:
                    lines.append(
                        "\nNote that Shrink Ray runs your script on a copy of the file "
                        "in a temporary directory. Here are the results of running it "
                        "in the current directory..."
                    )
                    other_exit_code = await self.run_script_on_file(
                        working=self.filename,
                        debug=True,
                        cwd=os.getcwd(),
                    )
                    # Include the output from running in current directory
                    if self._last_debug_output:
                        lines.append(self._last_debug_output)
                    if other_exit_code != local_exit_code:
                        lines.append(
                            f"This interestingness test is probably flaky as the first "
                            f"time we reran it locally it exited with {local_exit_code}, "
                            f"but the second time it exited with {other_exit_code}. "
                            "Please make sure your interestingness test is deterministic."
                        )
                    else:
                        lines.append(
                            "This suggests that your script depends on being run from "
                            "the current working directory. Please fix it to be "
                            "directory independent."
                        )
            else:
                assert self.initial_exit_code not in (None, 0)
                lines.append(
                    f"This exited with code 0, but previously the script exited with "
                    f"{self.initial_exit_code}. This suggests your interestingness "
                    "test exhibits nondeterministic behaviour."
                )

        return "\n".join(lines)

    async def report_error(self, e):
        error_message = await self.build_error_message(e)
        print(error_message, file=sys.stderr)
        sys.exit(1)

    def check_trivial_result(self, problem) -> str | None:
        """Check if the result is trivially small and return error message if so.

        Returns None if the result is acceptable, or an error message string
        if the result is trivial and trivial_is_error is True.
        """
        if len(problem.current_test_case) <= 1 and self.trivial_is_error:
            return (
                f"Reduced to a trivial test case of size {len(problem.current_test_case)}\n"
                "This probably wasn't what you intended. If so, please modify your "
                "interestingness test to be more restrictive.\n"
                "If you intended this behaviour, you can run with '--trivial-is-not-error' "
                "to suppress this message."
            )
        return None


@define(slots=False)
class ShrinkRayStateSingleFile(ShrinkRayState[bytes]):
    def new_reducer(self, problem: ReductionProblem[bytes]) -> Reducer[bytes]:
        return ShrinkRay(problem, clang_delta=self.clang_delta_executable)

    def setup_formatter(self):
        from shrinkray.formatting import (
            default_reformat_data,
            determine_formatter_command,
        )

        if self.formatter.lower() == "none":

            async def format_data(test_case: bytes) -> bytes | None:
                await trio.lowlevel.checkpoint()
                return test_case

            self.can_format = False

        else:
            formatter_command = determine_formatter_command(
                self.formatter, self.filename
            )
            if formatter_command is not None:
                self.formatter_command = formatter_command

                async def format_data(test_case: bytes) -> bytes | None:
                    result = await self.run_formatter_command(
                        formatter_command, test_case
                    )
                    if result.returncode != 0:
                        return None
                    return result.stdout

            else:

                async def format_data(test_case: bytes) -> bytes | None:
                    await trio.lowlevel.checkpoint()
                    return default_reformat_data(test_case)

        self.__format_data = format_data

    async def format_data(self, test_case: bytes) -> bytes | None:
        return await self.__format_data(test_case)

    async def run_formatter_command(
        self, command: str | list[str], input: bytes
    ) -> subprocess.CompletedProcess:
        return await trio.run_process(
            command,
            stdin=input,
            capture_stdout=True,
            capture_stderr=True,
            check=False,
        )

    async def write_test_case_to_file_impl(self, working: str, test_case: bytes):
        async with await trio.open_file(working, "wb") as o:
            await o.write(test_case)

    async def is_interesting(self, test_case: bytes) -> bool:
        async with self.is_interesting_limiter:
            return await self.run_for_exit_code(test_case) == 0

    async def print_exit_message(self, problem):
        formatting_increase = 0
        final_result = problem.current_test_case
        reformatted = await self.attempt_format(final_result)
        if reformatted != final_result:
            # attempt_format only returns a different value if is_interesting was True
            async with await trio.open_file(self.filename, "wb") as o:
                await o.write(reformatted)
            formatting_increase = max(0, len(reformatted) - len(final_result))
            final_result = reformatted

        if len(problem.current_test_case) <= 1 and self.trivial_is_error:
            print(
                f"Reduced to a trivial test case of size {len(problem.current_test_case)}"
            )
            print(
                "This probably wasn't what you intended. If so, please modify your interestingness test "
                "to be more restrictive.\n"
                "If you intended this behaviour, you can run with '--trivial-is-not-error' to "
                "suppress this message."
            )
            sys.exit(1)

        else:
            print("Reduction completed!")
            stats = problem.stats
            if self.initial == final_result:
                print("Test case was already maximally reduced.")
            elif len(final_result) < len(self.initial):
                print(
                    f"Deleted {humanize.naturalsize(stats.initial_test_case_size - len(final_result))} "
                    f"out of {humanize.naturalsize(stats.initial_test_case_size)} "
                    f"({(1.0 - len(final_result) / stats.initial_test_case_size) * 100:.2f}% reduction) "
                    f"in {humanize.precisedelta(timedelta(seconds=time.time() - stats.start_time))}"
                )
            elif len(final_result) == len(self.initial):
                print("Some changes were made but no bytes were deleted")
            else:
                print(
                    f"Running reformatting resulted in an increase of {humanize.naturalsize(formatting_increase)}."
                )


class ShrinkRayDirectoryState(ShrinkRayState[dict[str, bytes]]):
    def setup_formatter(self): ...

    @property
    def extra_problem_kwargs(self):
        def dict_size(test_case: dict[str, bytes]) -> int:
            return sum(len(v) for v in test_case.values())

        def dict_sort_key(test_case: dict[str, bytes]) -> Any:
            return (
                len(test_case),
                dict_size(test_case),
                sorted((k, shortlex(v)) for k, v in test_case.items()),
            )

        return {
            "sort_key": dict_sort_key,
            "size": dict_size,
        }

    def new_reducer(
        self, problem: ReductionProblem[dict[str, bytes]]
    ) -> Reducer[dict[str, bytes]]:
        return DirectoryShrinkRay(
            target=problem, clang_delta=self.clang_delta_executable
        )

    async def write_test_case_to_file_impl(
        self, working: str, test_case: dict[str, bytes]
    ):
        shutil.rmtree(working, ignore_errors=True)
        os.makedirs(working, exist_ok=True)
        for k, v in test_case.items():
            f = os.path.join(working, k)
            os.makedirs(os.path.dirname(f), exist_ok=True)
            async with await trio.open_file(f, "wb") as o:
                await o.write(v)

    async def format_data(self, test_case: dict[str, bytes]) -> dict[str, bytes] | None:
        # Formatting not supported for directory reduction
        return None

    async def run_formatter_command(
        self, command: str | list[str], input: dict[str, bytes]
    ) -> subprocess.CompletedProcess:
        # Formatting not supported for directory reduction
        raise NotImplementedError("Directory formatting not supported")

    async def print_exit_message(self, problem):
        print("All done!")



"""Worker subprocess that runs the reducer with trio and communicates via JSON protocol."""

import os
import sys
import time
import traceback
from contextlib import aclosing
from typing import Any, Protocol

import trio
from binaryornot.helpers import is_binary_string

from shrinkray.problem import InvalidInitialExample
from shrinkray.subprocess.protocol import (
    PassStatsData,
    ProgressUpdate,
    Request,
    Response,
    deserialize,
    serialize,
)


class InputStream(Protocol):
    """Protocol for async input streams."""

    def __aiter__(self) -> "InputStream": ...
    async def __anext__(self) -> bytes | bytearray: ...
    async def aclose(self) -> None: ...


class OutputStream(Protocol):
    """Protocol for output streams."""

    async def send(self, data: bytes) -> None: ...


class StdoutStream:
    """Wrapper around sys.stdout for the OutputStream protocol."""

    async def send(self, data: bytes) -> None:
        sys.stdout.write(data.decode("utf-8"))
        sys.stdout.flush()


class ReducerWorker:
    """Runs the reducer in a subprocess with JSON protocol communication."""

    def __init__(
        self,
        input_stream: InputStream | None = None,
        output_stream: OutputStream | None = None,
    ):
        self.running = False
        self.reducer = None
        self.problem = None
        self.state = None
        self._cancel_scope: trio.CancelScope | None = None
        # Parallelism tracking
        self._parallel_samples = 0
        self._parallel_total = 0
        # I/O streams - None means use stdin/stdout
        self._input_stream = input_stream
        self._output_stream = output_stream

    async def emit(self, msg: Response | ProgressUpdate) -> None:
        """Write a message to the output stream."""
        line = serialize(msg) + "\n"
        if self._output_stream is not None:
            await self._output_stream.send(line.encode("utf-8"))
        else:
            sys.stdout.write(line)
            sys.stdout.flush()

    async def read_commands(
        self,
        input_stream: InputStream | None = None,
        task_status: trio.TaskStatus[None] = trio.TASK_STATUS_IGNORED,
    ) -> None:
        """Read commands from input stream and dispatch them."""
        task_status.started()

        # Use provided stream, or instance stream, or default to stdin
        stream: InputStream
        if input_stream is not None:
            stream = input_stream
        elif self._input_stream is not None:
            stream = self._input_stream
        else:
            stream = trio.lowlevel.FdStream(os.dup(sys.stdin.fileno()))

        buffer = b""
        async with aclosing(stream) as aiter:
            async for chunk in aiter:
                buffer += chunk
                while b"\n" in buffer:
                    line, buffer = buffer.split(b"\n", 1)
                    if line:
                        await self.handle_line(line.decode("utf-8"))

    async def handle_line(self, line: str) -> None:
        """Handle a single command line."""
        try:
            request = deserialize(line)
            if not isinstance(request, Request):
                await self.emit(Response(id="", error="Expected a request"))
                return
            response = await self.handle_command(request)
            await self.emit(response)
        except Exception as e:
            traceback.print_exc()
            await self.emit(Response(id="", error=str(e)))

    async def handle_command(self, request: Request) -> Response:
        """Handle a command request and return a response."""
        match request.command:
            case "start":
                return await self._handle_start(request.id, request.params)
            case "status":
                return self._handle_status(request.id)
            case "cancel":
                return self._handle_cancel(request.id)
            case "disable_pass":
                return self._handle_disable_pass(request.id, request.params)
            case "enable_pass":
                return self._handle_enable_pass(request.id, request.params)
            case "skip_pass":
                return self._handle_skip_pass(request.id)
            case _:
                return Response(
                    id=request.id, error=f"Unknown command: {request.command}"
                )

    async def _handle_start(self, request_id: str, params: dict) -> Response:
        """Start the reduction process."""
        if self.running:
            return Response(id=request_id, error="Already running")

        try:
            await self._start_reduction(params)
            return Response(id=request_id, result={"status": "started"})
        except* InvalidInitialExample as excs:
            assert len(excs.exceptions) == 1
            (e,) = excs.exceptions
            # Build a detailed error message for invalid initial examples
            if self.state is not None:
                error_message = await self.state.build_error_message(e)
            else:
                error_message = str(e)
        except* Exception as e:
            traceback.print_exc()
            error_message = str(e.exceptions[0])
        return Response(id=request_id, error=error_message)

    async def _start_reduction(self, params: dict) -> None:
        """Initialize and start the reduction."""
        from shrinkray.cli import InputType
        from shrinkray.passes.clangdelta import (
            C_FILE_EXTENSIONS,
            ClangDelta,
            find_clang_delta,
        )
        from shrinkray.state import (
            ShrinkRayDirectoryState,
            ShrinkRayStateSingleFile,
        )
        from shrinkray.work import Volume

        filename = params["file_path"]
        test = params["test"]
        parallelism = params.get("parallelism", os.cpu_count() or 1)
        timeout = params.get("timeout")  # None means dynamic timeout
        seed = params.get("seed", 0)
        input_type = InputType[params.get("input_type", "all")]
        in_place = params.get("in_place", False)
        formatter = params.get("formatter", "default")
        volume = Volume[params.get("volume", "normal")]
        no_clang_delta = params.get("no_clang_delta", False)
        clang_delta_path = params.get("clang_delta", "")
        trivial_is_error = params.get("trivial_is_error", True)

        clang_delta_executable = None
        if os.path.splitext(filename)[1] in C_FILE_EXTENSIONS and not no_clang_delta:
            if not clang_delta_path:
                clang_delta_path = find_clang_delta()
            if clang_delta_path:
                clang_delta_executable = ClangDelta(clang_delta_path)

        state_kwargs: dict[str, Any] = {
            "input_type": input_type,
            "in_place": in_place,
            "test": test,
            "timeout": timeout,
            "base": os.path.basename(filename),
            "parallelism": parallelism,
            "filename": filename,
            "formatter": formatter,
            "trivial_is_error": trivial_is_error,
            "seed": seed,
            "volume": volume,
            "clang_delta_executable": clang_delta_executable,
        }

        if os.path.isdir(filename):
            files = [os.path.join(d, f) for d, _, fs in os.walk(filename) for f in fs]
            initial = {}
            for f in files:
                with open(f, "rb") as i:
                    initial[os.path.relpath(f, filename)] = i.read()
            self.state = ShrinkRayDirectoryState(initial=initial, **state_kwargs)
        else:
            with open(filename, "rb") as reader:
                initial = reader.read()
            self.state = ShrinkRayStateSingleFile(initial=initial, **state_kwargs)

        self.problem = self.state.problem
        self.reducer = self.state.reducer

        # Validate initial example before starting - this will raise
        # InvalidInitialExample if the initial test case fails
        await self.problem.setup()

        self.running = True

    def _handle_status(self, request_id: str) -> Response:
        """Get current status."""
        if not self.running or self.problem is None:
            return Response(id=request_id, result={"running": False})

        stats = self.problem.stats
        return Response(
            id=request_id,
            result={
                "running": True,
                "status": self.reducer.status if self.reducer else "",
                "size": stats.current_test_case_size,
                "original_size": stats.initial_test_case_size,
                "calls": stats.calls,
                "reductions": stats.reductions,
            },
        )

    def _handle_cancel(self, request_id: str) -> Response:
        """Cancel the reduction."""
        if self._cancel_scope is not None:
            self._cancel_scope.cancel()
        self.running = False
        return Response(id=request_id, result={"status": "cancelled"})

    def _get_known_pass_names(self) -> set[str]:
        """Get the set of known pass names from pass stats."""
        if self.reducer is None or self.reducer.pass_stats is None:
            return set()
        return set(self.reducer.pass_stats._stats.keys())

    def _handle_disable_pass(self, request_id: str, params: dict) -> Response:
        """Disable a reduction pass by name."""
        pass_name = params.get("pass_name", "")
        if not pass_name:
            return Response(id=request_id, error="pass_name is required")

        known_passes = self._get_known_pass_names()
        if known_passes and pass_name not in known_passes:
            return Response(
                id=request_id,
                error=f"Unknown pass '{pass_name}'. Known passes: {sorted(known_passes)}",
            )

        if self.reducer is not None and hasattr(self.reducer, "disable_pass"):
            self.reducer.disable_pass(pass_name)
            return Response(
                id=request_id, result={"status": "disabled", "pass_name": pass_name}
            )
        return Response(id=request_id, error="Reducer does not support pass control")

    def _handle_enable_pass(self, request_id: str, params: dict) -> Response:
        """Enable a previously disabled reduction pass."""
        pass_name = params.get("pass_name", "")
        if not pass_name:
            return Response(id=request_id, error="pass_name is required")

        known_passes = self._get_known_pass_names()
        if known_passes and pass_name not in known_passes:
            return Response(
                id=request_id,
                error=f"Unknown pass '{pass_name}'. Known passes: {sorted(known_passes)}",
            )

        if self.reducer is not None and hasattr(self.reducer, "enable_pass"):
            self.reducer.enable_pass(pass_name)
            return Response(
                id=request_id, result={"status": "enabled", "pass_name": pass_name}
            )
        return Response(id=request_id, error="Reducer does not support pass control")

    def _handle_skip_pass(self, request_id: str) -> Response:
        """Skip the currently running pass."""
        if self.reducer is not None and hasattr(self.reducer, "skip_current_pass"):
            self.reducer.skip_current_pass()
            return Response(id=request_id, result={"status": "skipped"})
        return Response(id=request_id, error="Reducer does not support pass control")

    def _get_content_preview(self) -> tuple[str, bool]:
        """Get a preview of the current test case content."""
        if self.problem is None:
            return "", False

        test_case = self.problem.current_test_case

        # Handle directory mode
        if isinstance(test_case, dict):
            lines = []
            for name, content in sorted(test_case.items()):
                size = len(content)
                lines.append(f"{name}: {size} bytes")
            return "\n".join(lines[:50]), False

        # Handle single file mode
        hex_mode = is_binary_string(
            test_case[:1024] if len(test_case) > 1024 else test_case
        )

        if hex_mode:
            # Show hex dump for binary files
            preview_bytes = test_case[:512]
            lines = []
            for i in range(0, len(preview_bytes), 16):
                chunk = preview_bytes[i : i + 16]
                hex_part = " ".join(f"{b:02x}" for b in chunk)
                ascii_part = "".join(chr(b) if 32 <= b < 127 else "." for b in chunk)
                lines.append(f"{i:08x}  {hex_part:<48}  {ascii_part}")
            return "\n".join(lines), True
        else:
            # Show text for text files
            # Send up to 100KB of content - the TUI will handle display truncation
            try:
                text = test_case.decode("utf-8", errors="replace")
                # Truncate by character count to handle files with very long lines
                max_chars = 100_000
                if len(text) > max_chars:
                    text = text[:max_chars]
                return text, False
            except Exception:
                return "", True

    async def _build_progress_update(self) -> ProgressUpdate | None:
        """Build a progress update from current state."""
        if self.problem is None:
            return None

        stats = self.problem.stats
        content_preview, hex_mode = self._get_content_preview()

        # Get parallel workers count and track average
        parallel_workers = 0
        if self.state is not None and hasattr(self.state, "parallel_tasks_running"):
            parallel_workers = self.state.parallel_tasks_running
            self._parallel_samples += 1
            self._parallel_total += parallel_workers

        # Calculate parallelism stats
        average_parallelism = 0.0
        effective_parallelism = 0.0
        if self._parallel_samples > 0:
            average_parallelism = self._parallel_total / self._parallel_samples
            wasteage = (
                stats.wasted_interesting_calls / stats.calls if stats.calls > 0 else 0.0
            )
            effective_parallelism = average_parallelism * (1.0 - wasteage)

        # Collect pass statistics in run order (only those with test evaluations)
        pass_stats_list = []
        current_pass_name = ""
        if self.reducer is not None:
            # Get the currently running pass name
            current_pass = self.reducer.current_reduction_pass
            if current_pass is not None:
                current_pass_name = getattr(current_pass, "__name__", "")

            # Get all stats in the order they were first run
            pass_stats = self.reducer.pass_stats
            if pass_stats is not None:
                all_stats = pass_stats.get_stats_in_order()

                # Only include passes that have made at least one test evaluation
                pass_stats_list = [
                    PassStatsData(
                        pass_name=ps.pass_name,
                        bytes_deleted=ps.bytes_deleted,
                        run_count=ps.run_count,
                        test_evaluations=ps.test_evaluations,
                        successful_reductions=ps.successful_reductions,
                        success_rate=ps.success_rate,
                    )
                    for ps in all_stats
                    if ps.test_evaluations > 0
                ]

        # Get disabled passes
        if self.reducer is not None and hasattr(self.reducer, "disabled_passes"):
            disabled_passes = list(self.reducer.disabled_passes)
        else:
            disabled_passes = []

        return ProgressUpdate(
            status=self.reducer.status if self.reducer else "",
            size=stats.current_test_case_size,
            original_size=stats.initial_test_case_size,
            calls=stats.calls,
            reductions=stats.reductions,
            interesting_calls=stats.interesting_calls,
            wasted_calls=stats.wasted_interesting_calls,
            runtime=time.time() - stats.start_time,
            parallel_workers=parallel_workers,
            average_parallelism=average_parallelism,
            effective_parallelism=effective_parallelism,
            time_since_last_reduction=stats.time_since_last_reduction(),
            content_preview=content_preview,
            hex_mode=hex_mode,
            pass_stats=pass_stats_list,
            current_pass_name=current_pass_name,
            disabled_passes=disabled_passes,
        )

    async def emit_progress_updates(self) -> None:
        """Periodically emit progress updates."""
        # Emit initial progress update immediately
        update = await self._build_progress_update()
        if update is not None:
            await self.emit(update)

        while self.running:
            await trio.sleep(0.1)
            update = await self._build_progress_update()
            if update is not None:
                await self.emit(update)

    async def run_reducer(self) -> None:
        """Run the reducer."""
        if self.reducer is None:
            return

        try:
            with trio.CancelScope() as scope:
                self._cancel_scope = scope
                await self.reducer.run()

            # Check for trivial result after successful completion
            if self.state is not None and self.problem is not None:
                trivial_error = self.state.check_trivial_result(self.problem)
                if trivial_error:
                    await self.emit(Response(id="", error=trivial_error))
        except* InvalidInitialExample as excs:
            assert len(excs.exceptions) == 1
            (e,) = excs.exceptions
            # Build a detailed error message for invalid initial examples
            if self.state is not None:
                error_message = await self.state.build_error_message(e)
            else:
                error_message = str(e)
            await self.emit(Response(id="", error=error_message))
        except* Exception as e:
            # Catch any other exception during reduction and emit as error
            traceback.print_exc()
            await self.emit(Response(id="", error=str(e.exceptions[0])))
        finally:
            self._cancel_scope = None
            self.running = False

    async def run(self) -> None:
        """Main entry point for the worker."""
        async with trio.open_nursery() as nursery:
            await nursery.start(self.read_commands)

            # Wait for start command
            while not self.running:
                await trio.sleep(0.01)

            # Start progress updates and reducer
            nursery.start_soon(self.emit_progress_updates)
            await self.run_reducer()

            # Emit final progress update before completion
            final_update = await self._build_progress_update()
            if final_update is not None:
                await self.emit(final_update)

            # Signal completion
            await self.emit(Response(id="", result={"status": "completed"}))
            nursery.cancel_scope.cancel()


def main() -> None:
    """Entry point for the worker subprocess."""
    worker = ReducerWorker()
    trio.run(worker.run)


if __name__ == "__main__":
    main()



from random import Random

import hypothesmith
import pytest
import trio
from hypothesis import Phase, assume, example, given, note, settings
from hypothesis import strategies as st
from hypothesis.errors import Frozen, StopTest

from shrinkray.passes.python import is_python
from shrinkray.problem import BasicReductionProblem, default_sort_key
from shrinkray.reducer import ShrinkRay
from shrinkray.work import Volume, WorkContext
from tests.helpers import assert_no_blockers, assert_reduces_to, direct_reductions


def tidy_python_example(s):
    results = []
    for line in s.splitlines():
        line, *_ = line.split("#")
        line = line.strip()
        if line:
            results.append(line)
    output = "\n".join(results)
    if output.startswith('"""'):
        output = output[3:]
        i = output.index('"""')
        output = output[i + 3 :]
    return output.strip() + "\n"


POTENTIAL_BLOCKERS = [
    b"\n",
    b"s",
    b"0",
    b"()",
    b"[]",
    b"\t\t",
    b"#\x00",
    b"#\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
    b"\t\t\t\t\t\t\t\t\t\t\t\t",
    # b"from\t\t\t.\t\t\t\t\t\timport\ta\t\t\t\t\t\t\t\t\t\t\nclass\t\to\t\t\t\t\t\t\t:\n\tdef\t\t\t\t\ta\t(\t\t\t\t\t\t\t\t\t)\t\t\t\t\t\t\t:...\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
    # "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t",
]

python_files = st.builds(
    lambda s, e: s.encode(e),
    hypothesmith.from_grammar(),
    st.sampled_from(("utf-8",)),
)
test_cases = (python_files).filter(lambda b: 1 < len(b) <= 1000)


common_settings = settings(deadline=None, max_examples=10, report_multiple_bugs=False)


@pytest.mark.slow
@common_settings
@given(
    initial=test_cases,
    rnd=st.randoms(use_true_random=True),
    is_interesting_sync=st.functions(
        returns=st.booleans(), pure=True, like=lambda test_case: False
    ),
    data=st.data(),
    parallelism=st.integers(1, 1),
)
async def test_can_shrink_arbitrary_problems(
    initial, rnd, data, parallelism, is_interesting_sync
):
    is_interesting_cache = {}

    initial_is_python = is_python(initial)

    async def is_interesting(test_case: bytes) -> bool:
        await trio.lowlevel.checkpoint()
        if test_case == initial:
            return True
        elif not test_case:
            return False
        try:
            return is_interesting_cache[test_case]
        except KeyError:
            pass
        if initial_is_python and not is_python(test_case):
            result = False
        else:
            try:
                result = is_interesting_sync(test_case)
            except (StopTest, Frozen):  # pragma: no cover
                result = False
        is_interesting_cache[test_case] = result
        if result:
            note(f"{test_case} is interesting")
        return result

    work = WorkContext(
        random=rnd,
        volume=Volume.quiet,
        parallelism=parallelism,
    )
    problem = BasicReductionProblem(
        initial=initial, is_interesting=is_interesting, work=work
    )

    reducer = ShrinkRay(problem)

    with trio.move_on_after(10) as cancel_scope:
        await reducer.run()
    assert not cancel_scope.cancelled_caught

    assert len(problem.current_test_case) <= len(initial)


@pytest.mark.slow
@settings(common_settings, phases=[Phase.explicit])
@example(b":\x80", 1)
@example(b"#\x80", 1)
@example(
    initial=b"..........................................................................................................................................................................................................................................................................................................................................................................................",
    parallelism=2,
)
@given(
    initial=test_cases,
    parallelism=st.integers(1, 10),
)
async def test_can_fail_to_shrink_arbitrary_problems(initial, parallelism):
    async def is_interesting(test_case: bytes) -> bool:
        await trio.lowlevel.checkpoint()
        return test_case == initial

    work = WorkContext(
        random=Random(0),
        volume=Volume.quiet,
        parallelism=parallelism,
    )
    problem = BasicReductionProblem(
        initial=initial, is_interesting=is_interesting, work=work
    )

    reducer = ShrinkRay(problem)

    with trio.move_on_after(10) as cancel_scope:
        await reducer.run()
    assert not cancel_scope.cancelled_caught

    assert problem.current_test_case == initial


@example(b"from p import*", 1)
@example(
    b'from types import MethodType\ndef is_hypothesis_test(test):\nif isinstance(test, MethodType):\nreturn is_hypothesis_test(test.__func__)\nreturn getattr(test, "is_hypothesis_test", False)',
    1,
)
@example(b"''", 1)
@example(b"# Hello world", 1)
@example(b"\x00\x01", 1)
@common_settings
@given(
    initial=test_cases,
    parallelism=st.integers(1, 10),
)
async def test_can_succeed_at_shrinking_arbitrary_problems(initial, parallelism):
    initial_is_python = is_python(initial)

    async def is_interesting(test_case: bytes) -> bool:
        if initial_is_python and not is_python(test_case):
            return False
        await trio.lowlevel.checkpoint()
        return len(test_case) > 0

    work = WorkContext(
        random=Random(0),
        volume=Volume.quiet,
        parallelism=parallelism,
    )
    problem = BasicReductionProblem(
        initial=initial, is_interesting=is_interesting, work=work
    )

    reducer = ShrinkRay(problem)

    with trio.move_on_after(10) as cancel_scope:
        await reducer.run()
    assert not cancel_scope.cancelled_caught

    assert len(problem.current_test_case) == 1


@pytest.mark.slow
def test_no_blockers():
    assert_no_blockers(
        potential_blockers=POTENTIAL_BLOCKERS,
        is_interesting=is_python,
        lower_bounds=[1, 2, 5, 10],
    )


@pytest.mark.skip(
    reason="Known issue: reducer doesn't consistently find same reductions"
)
@common_settings
@given(st.binary(), st.data())
def test_always_reduces_to_each_direct_reduction(origin, data):
    reductions = sorted(direct_reductions(origin), key=default_sort_key, reverse=True)

    assume(reductions)

    target = data.draw(st.sampled_from(reductions))

    assert_reduces_to(origin=origin, target=target, language_restrictions=False)


@pytest.mark.skip(reason="Known issue: parallelism can prevent some reductions")
@common_settings
@given(st.binary(), st.integers(2, 8), st.data())
def test_parallelism_never_prevents_reduction(origin, parallelism, data):
    reductions = sorted(direct_reductions(origin), key=default_sort_key, reverse=True)

    assume(reductions)

    parallel_reductions = sorted(
        direct_reductions(origin, parallelism=parallelism),
        key=default_sort_key,
        reverse=True,
    )

    assert set(reductions).issubset(parallel_reductions)

    target = data.draw(st.sampled_from(reductions))

    assert_reduces_to(
        origin=origin,
        target=target,
        parallelism=parallelism,
        language_restrictions=False,
    )



"""Subprocess communication for separating reducer from UI."""

from shrinkray.subprocess.client import SubprocessClient
from shrinkray.subprocess.protocol import (
    ProgressUpdate,
    Request,
    Response,
    decode_bytes,
    deserialize,
    encode_bytes,
    serialize,
)


__all__ = [
    "Request",
    "Response",
    "ProgressUpdate",
    "serialize",
    "deserialize",
    "encode_bytes",
    "decode_bytes",
    "SubprocessClient",
]
